/*
 * @(#)Tracer.java    Pzt 10 May 2010 19:09:18 EEST
 *
 * Murat iNAN, dilmerkezi.net
 */
/*
*** This program wrote under GNU/General Public License ***

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
*/
/*
 * Thanks for Timothy J. Hickey (cs.brandeis.edu).
 * We learnt ray tracing from HE.
 */
package net.dilmerkezi.main;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;

import java.io.*;

import javax.imageio.*;

import javax.swing.*;

//custom
import net.dilmerkezi.point.*;
import net.dilmerkezi.light.*;
import net.dilmerkezi.ray.*;
import net.dilmerkezi.object.*;

/**
 * @author Murat iNAN, dilmerkezi.net
 */
public final class Tracer
           extends JPanel
        implements ActionListener
{

    /** Predefined width*/
    private int width = 256;
    
    /** Predefined height*/
    private int height = 256;

    /** Width field*/
    private final JTextField widthField = new JTextField ("256");

    /** Height field*/
    private final JTextField heightField = new JTextField ("256");

    /** Eye point*/
    private Point3D eyepoint = new Vector3D (0, 0, 0);
    
    /** Eye field*/
    private final JTextField eyeField = new JTextField ("0;0;0");
    
    /** Floor*/
    private Floor3D floor = new Floor3D (2.5);
    
    /** Floor field*/
    private final JTextField floorField = new JTextField ("2");
    
    /** Floor modules*/
    private int floorModules = 0x4;
    
    /** F modules count field*/
    private final JTextField modulesField = new JTextField ("6");
    
    /** Light*/
    private Light3D light = 
        new Light3D (new Vector3D (2.0, 4.0, 1.0), Color.white);
    
    /** Light field*/
    private final JTextField lightField = new JTextField ("2.0;6.0;1.0");
    
    /** Light color*/
    private Color lightColor = Color.white;
    
    /** Light color button*/
    private final JButton lightColorButton = new JButton ("Light color");
    
    /** Predefined color*/
    private Color predefinedColor = Color.blue;
    
    /** Predefined color button*/
    private final JButton preColorButton = new JButton ("PreColor");
    
    /** Tracing ray depth*/
    private int depth = 1;
    
    /** Depth field*/
    private final JTextField depthField = new JTextField ("1");
    
    /** Image for saving*/
    private BufferedImage bimg = null;
    
    /** Npanel*/
    private JPanel npanel = new JPanel (new GridLayout (6, 4, 0, 0));
    
    /** Center panel*/
    private JPanel cpanel = new XPanel ();
    
    /** File selection tool*/
    private final JFileChooser jfc = new JFileChooser ();
    

    /** Singleton constructor*/
    private Tracer ()
    {
        super (new BorderLayout ());
        
        depthField.setEditable (false);
        
        npanel.add (new JLabel ("Width: ", JLabel.LEFT));
        npanel.add (widthField);
        
        npanel.add (new JLabel ("Height: ", JLabel.LEFT));
        npanel.add (heightField);
        
        npanel.add (new JLabel ("Eye: ", JLabel.LEFT));
        npanel.add (eyeField);
        
        npanel.add (new JLabel ("Floor: ", JLabel.LEFT));
        npanel.add (floorField);
        
        npanel.add (new JLabel ("Modules: ", JLabel.LEFT));
        npanel.add (modulesField);
        
        npanel.add (new JLabel ("Depth: ", JLabel.LEFT));
        npanel.add (depthField);
        
        npanel.add (new JLabel ("Light: ", JLabel.LEFT));
        npanel.add (lightField);
        
        npanel.add (new JLabel ("Light Color: ", JLabel.LEFT));
        lightColorButton.setBackground (Color.white);
        lightColorButton.addActionListener (new ActionListener ()
        {
            public void actionPerformed (ActionEvent evt)
            {
                Color c = JColorChooser.showDialog (cpanel, "Light Color", Color.white);
                if (c != null)
                {
                    lightColor = c;
                    lightColorButton.setBackground (lightColor);
                    return;
                }
            }
        });
        npanel.add (lightColorButton);
        
        npanel.add (new JLabel ("PreColor: : ", JLabel.LEFT));
        preColorButton.setBackground (Color.blue);
        preColorButton.addActionListener (new ActionListener ()
        {
            public void actionPerformed (ActionEvent evt)
            {
                Color c = JColorChooser.showDialog (cpanel, "Pre Color", Color.white);
                if (c != null)
                {
                    predefinedColor = c;
                    preColorButton.setBackground (predefinedColor);
                    return;
                }
            }
        });
        npanel.add (preColorButton);
        
        JButton gbtn = new JButton ("Generate");
        gbtn.addActionListener (this);
        npanel.add (gbtn);

        JButton sbtn = new JButton ("Save");
        sbtn.addActionListener (new ActionListener ()
        {
            public void actionPerformed (ActionEvent evt)
            {
                try
                {
                    saveImage ();
                }
                catch (Exception e)
                {
                    JOptionPane.showMessageDialog (cpanel,
                     "Error occurred when saving file!",
                     "Error",
                     JOptionPane.ERROR_MESSAGE);
                    return;
                }
            }
        });
        npanel.add (sbtn);
        
        JButton cbtn = new JButton ("Clear");
        cbtn.addActionListener (new ActionListener ()
        {
            public void actionPerformed (ActionEvent evt)
            {
                bimg = null;
                cpanel.repaint ();
            }
        });
        npanel.add (cbtn);
        
        JButton vsButton = new JButton ("Visit");
        vsButton.addActionListener (new ActionListener ()
        {
            public void actionPerformed (ActionEvent evt)
            {
                int k = JOptionPane.showConfirmDialog (npanel, "Do you want to visit dilmerkezi.net/eng.html?");
                if (k != 0) return;
                String CMD = "";
                String os = System.getProperty ("os.name");
                if (os.indexOf ("Windows") != -1)
                {
                    CMD = "explorer http://www.dilmerkezi.net/eng.html";
                }
                else
                {
                    CMD = "iceweasel http://www.dilmerkezi.net/eng.html";
                }
                
                try
                {
                    Runtime rt = Runtime.getRuntime ();
                    Process p = rt.exec (CMD);
                    p.waitFor ();
                }
                catch (Exception e)
                {}
            }
        });
        npanel.add (vsButton);
        
        JButton ebtn = new JButton ("Exit");
        ebtn.addActionListener (new ActionListener ()
        {
            public void actionPerformed (ActionEvent evt)
            {
                System.exit (0);
            }
        });
        npanel.add (ebtn);
        
        JSplitPane split = new JSplitPane (
            JSplitPane.VERTICAL_SPLIT, npanel, new JScrollPane (cpanel));
        split.setOneTouchExpandable (true);
        add ("Center", split);
    }
    
    /**
     * Inner panel
     */
    private final class XPanel
                extends JPanel
    {
    
        /** Inner singleton*/
        private XPanel ()
        {
            super (true);
            setPreferredSize (new Dimension (512, 512));
        }
        
        /** @paramg Graphics object*/
        public void update (Graphics g)
        {
            paintComponent (g);
        }
        
        /** @paramg Graphics object*/
        protected void paintComponent (Graphics g)
        {
            super.paintComponent (g);
            
            Dimension d = this.getSize ();
            g.setColor (Color.white);
            
            g.fillRect (0, 0, d.width, d.height);
            
            if (bimg != null)
            {
                g.drawImage (bimg, 0, 0, this);
            }
        }
        
    }
    /** @param evt Action event object //generate*/
    public void actionPerformed (ActionEvent evt)
    {
        Thread t = new Thread ()
        {
            public void run ()
            {
                try
                {
                    generate ();
                }
                catch (Exception e)
                {
                    JOptionPane.showMessageDialog (cpanel,
                     "Error occurred when tracing ray!",
                     "Error",
                     JOptionPane.ERROR_MESSAGE);
                    return;
                }
            }
        };
        t.start ();
    }
    
    /** Generate*/
    private final void generate ()
    throws Exception
    {
        width = Integer.parseInt (widthField.getText ());
        height = Integer.parseInt (heightField.getText ());
        
        String s = eyeField.getText ();
        String eyes [] = s.split (";");
        eyepoint = new Vector3D (
            Double.parseDouble (eyes [0]),
            Double.parseDouble (eyes [1]),
            Double.parseDouble (eyes [2]));
        
        floorModules = Integer.parseInt (modulesField.getText ());
        
        if (floorModules > 6) floorModules = 6;
        
        floor = new Floor3D (
          Double.parseDouble (floorField.getText ()),
          floorModules<1?1:floorModules
          );
        
        lightColor = lightColorButton.getBackground ();
        eyes = (lightField.getText ()).split (";");
        light = new Light3D (
            new Vector3D (
            Double.parseDouble (eyes [0]),
            Double.parseDouble (eyes [1]),
            Double.parseDouble (eyes [2])), lightColor);
            
        depth = Integer.parseInt (depthField.getText ());
        predefinedColor = preColorButton.getBackground ();        
        
        bimg = getImage ();
        if (bimg == null) return;
        
        cpanel.repaint ();        
    }
    
    /**
     * @param ray Ray3D
     * @return A color
     */
    private final Color getColor (Ray3D ray)
    {
        double t = floor.rayIntersect (ray);
        if (t<=0) 
        {
            return predefinedColor;
        }
        
        Point3D p = ray.atTime (t);
        Color bg = floor.background (p);
        
        double intensity = 0.0;
        Point3D n = floor.normal (p);
        Point3D lvec = (light.getLightOriginPoint3D ()).subtract (p);
        double dist = lvec.length ();
        
        Ray3D rr = new Ray3D (p, lvec);
        double ts = floor.rayIntersect (rr);
        
        if (ts < 0)
        {
            double diff = light.diffuse (p, n);
            double spec = light.specular (p, n, eyepoint);
            intensity += (diff+spec);
        }
        
        int red = adjust (intensity*bg.getRed ()+(light.getLightColor().getRed ()/2));
        int grn = adjust (intensity*bg.getGreen ()+(light.getLightColor().getGreen ()/2));
        int blu = adjust (intensity*bg.getBlue ()+(light.getLightColor().getBlue ()/2));

        return new Color (red, grn, blu);        
    }

    /**
     * @param d Color value like double
     * @return A val between 0-255
     */
    private static final int adjust (double d)
    {
        int k=(int)d;
        if (k<0) return 0;
        if (k>255) return 255;
        return k;
    }
    
    /** Select image w-h ........ */
    private final BufferedImage getImage ()
    {
        if  (width > 512)  width = 512;
        if (height > 512) height = 512;
        
        BufferedImage bi = new BufferedImage (width, height, 1);
        Graphics2D g2d = bi.createGraphics ();
        
        g2d.setBackground (Color.white);
        g2d.clearRect (0, 0, width, height);
        
        for (int i=0; i<width; i++)
        {
            for (int j=0; j<height; j++)
            {
                double u = (double)(i-(width/10));
                double v = (double)(j-(height/4));
                
                Point3D npoint = new Vector3D (u, v, eyepoint.getZ ());
                Ray3D ray = new Ray3D (eyepoint, npoint);
                
                if (depth > 16) depth = 16;
                if (depth <= 1)
                {
                    Color c = getColor (ray);
                    g2d.setColor (c);
                    g2d.fillRect (i, j, 1, 1);
                }
                else
                {
                    Color cx = Color.black;
                    for (int m=1; m<depth; m++)
                    {
                    }
                }
            }
        };
        
        if (bi != null) bimg = bi;
        return bimg;
    }
    
    /** Saves image*/
    private final void saveImage ()
    throws Exception
    {
        if (bimg == null) return;
        int rep = jfc.showSaveDialog (cpanel);
        if (rep != JFileChooser.APPROVE_OPTION) return;
        File f = jfc.getSelectedFile ();
        ImageIO.write (bimg, "png", f);
        JOptionPane.showMessageDialog (cpanel, "Saved: "+f.getName ()+"");
    }
    
    /** Shows screen*/
    private static final void showScreen ()
    {
        try
        {
            UIManager.setLookAndFeel (
                "com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");
        }
        catch (Exception e)
        {}
        
        JFrame f = new JFrame ("Simple Ray Tracer, dilmerkezi.net/eng.html");
        f.setDefaultCloseOperation (JFrame.EXIT_ON_CLOSE);
        
        JComponent c = new Tracer ();
        c.setOpaque (true);
        f.setContentPane (c);
        
        f.pack ();
        f.setSize (640, 480);
        //f.setResizable (false);
        f.setLocationRelativeTo (null);
        f.setVisible (true);
    }
    
    /** @param args Arguments. The first method of applications*/
    final
    public static void main (String [] args)
    throws Exception
    {
        Runnable r = new Runnable ()
        {
            public void run ()
            {
                showScreen ();
            }
        };
        javax.swing.SwingUtilities.invokeLater (r);
    }
    
}
